/* $Id: vircam_opm.c,v 1.2 2010-09-09 12:11:09 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2005 Cambridge Astronomy Survey Unit
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2010-09-09 12:11:09 $
 * $Revision: 1.2 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "vircam_mods.h"
#include "vircam_utils.h"
#include "vircam_fits.h"
#include "vircam_mask.h"
#include "catalogue/imcore.h"

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Generate an object mask from an input image

    \par Name:
        vircam_opm
    \par Purpose:
        Generate an object mask from an input image
    \par Description:
        A frame and its confidence map are given. Detection thresholds and
	various other parameters are also given. Output is an object mask of
	the extracted objects. The result is written to the input cpl_image
	bad pixel mask.
    \par Language:
        C
    \param infile
        The input frame with the image to be analysed
    \param conf
        The input frame with the confidence map
    \param ipix
        The minimum allowable size of an object
    \param threshold
        The detection threshold in sigma above sky
    \param nbsize
        The smoothing box size for background map estimation
    \param filtfwhm
        The FWHM of the smoothing kernel in the detection algorithm
    \param niter
        The number of detection iterations to be done
    \param status
        The input/output status that has the same return value a below
    \retval VIR_OK 
        if everything is ok
    \retval VIR_WARN,VIR_FATAL
        errors in the called routines
    \par QC headers:
        The following values will go into the extension propertylist
	- \b MEAN_SKY
	    Mean sky brightness in ADU
	- \b SKY_NOISE
	    Pixel noise at sky level in ADU
    \author
        Jim Lewis, CASU

 */
/*---------------------------------------------------------------------------*/

extern int vircam_opm(vir_fits *infile, vir_fits *conf, int ipix,
		      float threshold, int nbsize, float filtfwhm, 
		      int niter, int *status) {
    int retval;
    cpl_image *im;
    vir_fits *in_copy,*c_copy;

    /* Inherited status */

    if (*status != VIR_OK)
	return(*status);

    /* Copy the input */

    im = vircam_fits_get_image(infile);
    in_copy = vircam_fits_duplicate(infile);
    c_copy = vircam_fits_duplicate(conf);

    /* Call the main processing routine and create the mask */

    retval = imcore_opm(in_copy,c_copy,ipix,threshold,nbsize,filtfwhm,niter);
    if (retval != VIR_OK) {
        vircam_fits_delete(in_copy);
        vircam_fits_delete(c_copy);
	FATAL_ERROR
    }	

    /* Copy the mask over */

    cpl_mask_or(cpl_image_get_bpm(im),
		cpl_image_get_bpm(vircam_fits_get_image(in_copy)));

    /* Ditch the copies */

    vircam_fits_delete(in_copy);
    vircam_fits_delete(c_copy);

    /* Get out of here */

    GOOD_STATUS
}

/**@}*/

/* 

$Log: not supported by cvs2svn $
Revision 1.1  2008/10/24 11:08:12  jim
New entry


*/
